/*
 * CExpat
 * Copyright (c) 2001 Robert Mooney (rjmooney@aboveground.cx), All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 */

#include "CExpat.h"


// Initialize a pointer to the expat class
// This allows expat to call C++ member functions, and is set upon initialization of the expat class
//
static CExpat *g_expat = NULL;


// Non-member function which allows expat to call the member function "ElementStart" indirectly
void _ElementStart( void *vUserData, const XML_Char *ntsElement, const XML_Char **ppszAttributes )
{
	assert( g_expat );
	g_expat->ElementStart( vUserData, ntsElement, ppszAttributes );
}


// Non-member function which allows expat to call member function "ElementData" indirectly
void _ElementData( void *vUserData, const XML_Char *ntsString, int nStringLen )
{
	assert( g_expat );  
	g_expat->ElementData( vUserData, ntsString, nStringLen);
}


// Non-member function which allows expat to call member function "ElementEnd" indirectly
void _ElementEnd( void *vUserData, const XML_Char *ntsElement )
{
	assert( g_expat );
	g_expat->ElementEnd( vUserData, ntsElement );
}


// Constructor
CExpat::CExpat()
{
	g_expat = this;

	// Construct a new XML parser
	m_parser_handle = XML_ParserCreate( NULL );

	// If the construction failed
	if ( !m_parser_handle )
	{
		throw "Could not allocate memory for parser"; 
	}

	// Set the handler for start and end tags
	XML_SetElementHandler( m_parser_handle, _ElementStart, _ElementEnd );

	// Set the handler for element data
	XML_SetCharacterDataHandler( m_parser_handle, _ElementData );
}


// Constructor overload
CExpat::CExpat( const char *szFilename )
{
	g_expat = this;

	// Construct a new XML parser
	m_parser_handle = XML_ParserCreate( NULL );

	// If the construction failed
	if ( !m_parser_handle )
	{
		throw "Could not allocate memory for parser"; 
	}

	// Set the handler for start and end tags
	XML_SetElementHandler( m_parser_handle, _ElementStart, _ElementEnd );

	// Set the handler for element data
	XML_SetCharacterDataHandler( m_parser_handle, _ElementData );

	// Allocate and initialize
	char szBuffer[ 8192 ];
	int done = 0;

	// Open the specified file
	ifstream from( szFilename );
	if ( from.rdstate() )  throw "Could not open input file.\n";

	// Read to EOF
	while ( !done )
	{
		// Read data from the input file; store the bytes read
		from.read( &szBuffer[0], sizeof( szBuffer ) );

		// If the end of file is encountered, we're done reading
		if ( from.eof() )  done = 1; 

		// If another error is encountered
		else if ( from.rdstate() )  throw "Could not read data from input file"; 

		// Parse the data read
		Parse( szBuffer, from.gcount(), done );
	}
}


// Constructor overload
CExpat::CExpat( const char *szBuffer, const int nBufferLen, const int nFinal )
{
	g_expat = this;

	// Construct a new XML parser
	m_parser_handle = XML_ParserCreate( NULL );

	// If the construction failed
	if ( !m_parser_handle )
	{
		throw "Could not allocate memory for parser"; 
	}

	// Set the handler for start and end tags
	XML_SetElementHandler( m_parser_handle, _ElementStart, _ElementEnd );

	// Set the handler for element data
	XML_SetCharacterDataHandler( m_parser_handle, _ElementData );

	// Parse the specified buffer
	Parse( szBuffer, nBufferLen, nFinal );
}


// Decontructor
CExpat::~CExpat()
{
	g_expat = NULL;
}


// The "real" "element start" function
// Called when an element tag is found
//
void CExpat::ElementStart( void *vUserData, const XML_Char *ntsElement, const XML_Char **ppszAttributes ) 
{
	// The path string
	string path;

	// Initialize the path with the path to the parent
	if ( !m_stack.empty() )	path = m_stack.top().path;

	// Add the new element name to the path
	path += "/";
	path += ntsElement;

	// Allocate and initialize an element
	element_data_t data;
	data.content = "";
	data.child_vector_index = -1;			// XXX make sure we don't insert too many

	// Populate the element's attribute list
	for ( int n = 0; ppszAttributes[n]; n += 2 )
	{
		data.attributes[ ppszAttributes[n] ] = ppszAttributes[ n + 1 ];
	}

	// Insert the current path/element into the map
	element_map_t::iterator itor = m_element_map.insert(
		pair< string, element_data_t >( path, data )
	);

	// If there is data on the stack, we must be a child.
	// Insert the itor in the parent vector.
	//
	if ( !m_stack.empty() )
	{
		// Grab the parent element data
		element_data_t *parent_data = &( ( *m_stack.top().itor ).second );

		// If a child vector index for the parent does not yet exist, create it,
		// and insert the child's itor.
		//
		if ( parent_data->child_vector_index == -1 )
		{
			// Allocate an child_itor_vector and store the current element itor
			// (i.e. the itor of the element we are about to push onto the
			// stack).
			//
			child_itor_vector_t child_itor_vector;
			child_itor_vector.insert( child_itor_vector.end(), itor );

			// Initialize the child vector index, and store the itor vector
			// at the end of the child vector
			//
			parent_data->child_vector_index = m_child_vector.size();
			m_child_vector.insert( m_child_vector.end(), child_itor_vector );
		}

		// Otherwise, grab the itor vector from the child vector, and store
		// the current element itor.
		//
		else
		{
			// Allocate an child_itor_vector pointer, retrieve the child vector
			// for the element currently on the stack (i.e. the parent to 
			// the element we are about to push onto the stack), and store
			// the current element itor (i.e. the itor of the element we are
			// about to push onto the stack).
			//
			child_itor_vector_t *child_itor_vector;
			child_itor_vector = &( m_child_vector[ parent_data->child_vector_index ] );
			child_itor_vector->insert( child_itor_vector->end(), itor );
		}
	}

	// Store the path and itor information on the stack
	stack_element_t stack_element;
	stack_element.itor = itor;
	stack_element.path = path;
	m_stack.push( stack_element );
}


// The "real" "element data" function
// Called when text is found within an element
//
void CExpat::ElementData( void *vUserData, const XML_Char *ntsString, int nStringLen )
{
	int pos = 0;
	string local;

	// Append the data to the local string
	local.append( ntsString, nStringLen );

	// Remove extraneous characters
	while (1)
	{
		// Determine the position of the first extraneous char
		pos = local.find_first_of( "\n\t", 0 );

		// Break if no extraneous char exists
		if ( pos == -1 )  break;

		// Remove the extraneous char
		local.erase( pos, 1 );
	}

	// Return if the string length is 0
	if ( local.length() == 0 )  return;

	// Grab the current element iterator
	element_map_t::iterator itor = m_stack.top().itor;

	// Grab the current element data and store the new value string
	element_data_t *element_data = &( ( *itor ).second );
	element_data->content = local;
}


// The "real" "element end" function
// Called when an element tag is found to end
//
void CExpat::ElementEnd( void *vUserData, const char *ntsElement )
{
	m_stack.pop();
}


// Remove an attribute from the specified element, returns TRUE on success
bool CExpat::EraseAttribute( element_map_t::iterator parent_itor, const XML_Char *szName )
{
	element_data_t *parent_data = & ( (*parent_itor).second );
	return parent_data->attributes.erase( szName ) ? true : false;
}


// Returns a pointer to a child iterator vector, so that the caller can cycle through the 
// list of children associated with the parent iterator specified.
//
// Returns NULL if no children exist.
//
child_itor_vector_t* CExpat::GetChildItorVector( element_map_t::const_iterator parent_itor )
{
	// Grab the parent element data
	const element_data_t *parent_data = & ( (*parent_itor).second );

	// If the element has no children, return
	if ( parent_data->child_vector_index == -1 )  return NULL;

	// Return a pointer to the child iterator vector
	return &m_child_vector[ parent_data->child_vector_index ];
}


// Insert an element as a child to parent_itor, using the specified element name, attributes and content
void CExpat::Insert( element_map_t::iterator parent_itor, const char *szElementName, element_attribute_map_t *amAttributes, const char *szContent )
{
	// Add the new element name to the parent path
	string path = (*parent_itor).first;
	path += "/";
	path += szElementName;

	// Allocate and initialize an element
	element_data_t data;
	data.child_vector_index = -1;			// XXX make sure we don't insert too many

	// If attributes were specified
	if ( amAttributes )
	{
		// Populate the element's attribute list
		for ( element_attribute_map_t::iterator itor = amAttributes->begin();
				itor != amAttributes->end();
				++itor )
		{
			data.attributes[ (*itor).first ] = (*itor).second;
		}
	}

	// Store the element's content value
	data.content = szContent ? szContent : "";

	// Insert the new path/element into the map
	element_map_t::iterator itor = m_element_map.insert(
		pair< string, element_data_t >( path, data )
	);


	// Grab the parent element data
	element_data_t *parent_data = &( (*parent_itor).second );

	// If a child vector index for the parent does not yet exist, create it,
	// and insert the child's itor.
	//
	if ( parent_data->child_vector_index == -1 )
	{
		// Allocate an child_itor_vector and store the current element itor
		// (i.e. the itor of the element we are about to push onto the
		// stack).
		//
		child_itor_vector_t child_itor_vector;
		child_itor_vector.insert( child_itor_vector.end(), itor );

		// Initialize the child vector index, and store the itor vector
		// at the end of the child vector
		//
		parent_data->child_vector_index = m_child_vector.size();
		m_child_vector.insert( m_child_vector.end(), child_itor_vector );
	}

	// Otherwise, grab the itor vector from the child vector, and store
	// the current element itor.
	//
	else
	{
		// Allocate an child_itor_vector pointer, retrieve the child vector
		// for the element currently on the stack (i.e. the parent to 
		// the element we are about to push onto the stack), and store
		// the current element itor (i.e. the itor of the element we are
		// about to push onto the stack).
		//
		child_itor_vector_t *child_itor_vector;
		child_itor_vector = &( m_child_vector[ parent_data->child_vector_index ] );
		child_itor_vector->insert( child_itor_vector->end(), itor );
	}	
}


// Insert an element at the root
void CExpat::Insert( const char *szElementName, element_attribute_map_t *amAttributes, const char *szContent )
{
	// Create the new element name
	string path = "/";
	path += szElementName;

	// Allocate and initialize an element
	element_data_t data;
	data.child_vector_index = -1;			// XXX make sure we don't insert too many

	// If attributes were specified
	if ( amAttributes )
	{
		// Populate the element's attribute list
		for ( element_attribute_map_t::iterator itor = amAttributes->begin();
				itor != amAttributes->end();
				++itor )
		{
			data.attributes[ (*itor).first ] = (*itor).second;
		}
	}

	// Store the element's content value
	data.content = szContent ? szContent : "";

	// Insert the new path/element into the map
	m_element_map.insert(
		pair< string, element_data_t >( path, data )
	);	
}


// Insert an attribute in the specified element
void CExpat::InsertAttribute( element_map_t::iterator parent_itor, const char *szName, const char *szValue )
{
	// Grab the parent element data
	element_data_t *parent_data = &( (*parent_itor).second );

	// Insert the new attribute
	parent_data->attributes[ szName ] = szValue;
}


// Parse a buffer containing XML data
// Final indicates this is the last buffer in the sequence (i.e., EOF)
//
void CExpat::Parse( const char *szBuffer, const int nBufferLen, const int nFinal )
{
	// Parse the specified buffer
	// If there is an error, create an error string and throw
	//
	if ( !XML_Parse( m_parser_handle, szBuffer, nBufferLen, nFinal ) )
	{
		string error = "Cannot parse line ";
		error += XML_GetCurrentLineNumber( m_parser_handle );
		error += ":\n";
		error += XML_ErrorString( XML_GetErrorCode( m_parser_handle ) );
		error += "\n";

		throw error.c_str();
	}
}


// Parse the specified file
void CExpat::Parse( const char *szFilename )
{
	// Allocate and initialize
	char szBuffer[ 8192 ];
	int done = 0;

	// Open the specified file
	ifstream from( szFilename );
	if ( from.rdstate() )  throw "Could not open input file.\n";

	// Read to EOF
	while ( !done )
	{
		// Read data from the input file; store the bytes read
		from.read( &szBuffer[0], sizeof( szBuffer ) );

		// If the end of file is encountered, we're done reading
		if ( from.eof() )  done = 1; 

		// If another error is encountered
		else if ( from.rdstate() )  throw "Could not read data from input file"; 

		// Parse the data read
		Parse( szBuffer, from.gcount(), done );
	}
}


// Return the content portion of an element
const char* CExpat::GetContent( element_map_t::iterator parent_itor )
{
	// Grab the parent element data
	element_data_t *parent_data = &( (*parent_itor).second );

	// Assign the new content value
	return parent_data->content.c_str();
}


// Set the content portion of an element
void CExpat::SetContent( element_map_t::iterator parent_itor, const char *szContent )
{
	// Grab the parent element data
	element_data_t *parent_data = &( (*parent_itor).second );

	// Assign the new content value
	parent_data->content = szContent;
}
